!#################################################################
!
! MODULE: emission_co2_ff 
! 
! PUBLIC TYPES:
!
! PUBLIC MEMBER FUNCTIONS:
!
!   emission_co2_ff_init
!   emission_co2_ff_done
!   emission_co2_ff_calc
!
! PUBLIC DATA MEMBERS:
!
! DESCRIPTION: 
!
!   Compute CO2 fossil fuel emissions in kgC/gridbox/mo.
!
! REVISION HISTORY: 
!
!   Andy Jacobson, Dec 2012
!     adapted from ODIAC emisisons code originally
!     written by Mike Trudeau
!
!### macros #####################################################
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
!
#include "tm5.inc"
!
!#################################################################

module emission_co2_ff
  
  use GO,                  only : gol, goErr, goPr
  use emission_common,     only : enkf_parms, field_1x1, EmisInputDir
  use emission_common,     only : nmembersloc

  implicit none

  character(len=*), parameter   :: mname = 'emission_co2_ff'

  private

  character(len=1024)           :: ff_prefix
  character(len=256)            :: ff_varname

  public                        :: emission_co2_ff_init, emission_co2_ff_done, emission_co2_ff_calc

  character(len=2048)           :: FFInputDir ! default is EmisInputDir

contains

  subroutine emission_co2_ff_init( status )

    use GO,                only : TrcFile, Init, Done, ReadRc
    use global_data,       only : rcfile

    implicit none

    character(len=*), parameter :: rname = mname//'/emission_co2_ff_init'

    integer, intent(out)        :: status
    type(TrcFile)               :: rcF

    call Init( rcF, rcfile, status )
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'ff.input.dir', FFInputDir, status, default = EmisInputDir)
    IF_ERROR_RETURN(status=1)
    call ReadRc( rcF, 'ff.ncfile.prefix', ff_prefix, status, ff_prefix)
    IF_ERROR_RETURN(status=1)
    call ReadRc( rcF, 'ff.ncfile.varname', ff_varname , status, 'fossil')
    IF_ERROR_RETURN(status=1)
    call Done( rcF, status )
    IF_NOTOK_RETURN(status=1)
    
    write (gol,'("[emission_co2_ff_init]	FF input files to be read from """,a,""",")') trim(FFInputDir)
    call goPr
    write (gol,'("[emission_co2_ff_init]	   with filename prefix: """,a,""",")') trim(ff_prefix)
    call goPr
    write (gol,'("[emission_co2_ff_init]	   and variable name:  """,a,""".")') trim(ff_varname)
    call goPr

    status = 0

  end subroutine emission_co2_ff_init

  subroutine emission_co2_ff_calc( parms, co2_ff, flux_updated, status )

    use dims,                              only : idate, newmonth, newsrun, nlon360, nlat180, sec_month, dxy11, sec_year
    use MDF,                               only : MDF_NETCDF
    use ReadFromFile_MDF,                  only : ReadFromFile
    use go_string,                         only : goNum2str

    implicit none
    
    character(len=*), parameter                 :: rname = mname//'/emission_co2_ff_calc'
    type(enkf_parms), dimension(:), intent(in)  :: parms ! not currently used as fossil fuel emissions are not optimized
    real*4, dimension(nlon360, nlat180, 12)       :: ncArr
    integer                                     :: i, j, t ! lon, lat, time
    type(field_1x1), dimension(:), intent(out)  :: co2_ff
    integer, intent(out)                        :: status
    logical, intent(out)                        :: flux_updated
    integer                                     :: imember
    real					:: month2year

 
    flux_updated = .false.

    if ( newmonth .or. newsrun ) then
 
      flux_updated = .true.

      call ReadFromFile(fname = trim(FFInputDir)//'/'//trim(ff_prefix)//'.'//trim(goNum2str(idate(1), '(i4.4)'))//'.nc', & 
                        ftype = MDF_NETCDF, vname = trim(ff_varname), &
                        buffer = ncArr, status = status)
      IF_NOTOK_RETURN(status=1)

      do j = 1, nlat180
        ncArr(:, j, idate(2)) = 12e-3 * ncArr(:, j, idate(2)) * sec_month ! mol m-2 s-1 to kgC m-2 mon-1
        ncArr(:, j, idate(2)) = ncArr(:, j, idate(2)) * dxy11(j) ! kgC m-2 mon-1 to kgC gridbox-1 mon-1
      enddo

      ! ensemble members assigned same fossil fuel flux
      month2year = sec_year/sec_month
      do imember=1,nmembersloc
        co2_ff(imember)%surf = ncArr(:, :, idate(2))

        write (gol,'("[emission_co2_ff_calc]	member ",i4,": fossil fuels emission global integral ",f12.4," PgC/yr.")') imember, &
              sum(co2_ff(imember)%surf * 1.0E-12 *month2year)
        call goPr
      end do

    end if

    status = 0

  end subroutine emission_co2_ff_calc

  subroutine emission_co2_ff_done(status)

    implicit none

    integer, intent(out)        :: status
    character(len=*), parameter :: rname = mname//'/emission_co2_ff_done'

    status=0

  end subroutine emission_co2_ff_done

end module emission_co2_ff
